/* XML output implementation for the C++ compiler.

   Brad King (brad.king@kitware.com)
   Kitware, Inc.   (www.kitware.com)

This file WILL HOPEFULLY BE part of GNU CC (eventually).

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


/* Organization of this source:
   The entry point is do_xml_output(), which is called from the end of
   finish_translation_unit() in semantics.c.
   
   xml_output_* functions are used to output tree nodes that require
   non-trivial processing (like RECORD_TYPE).  This includes any nodes
   whose XML elements have beginning and ending tags (not empty elements)
   and have other elements nested inside them.

   print_*_begin_tag functions are used to write out the XML begin tags
   with their attributes.  The functions are responsible for extracting the
   details from a tree node needed specifically for the XML attributes on
   the begin tag.

   print_*_end_tag functions print the corresponding XML end tags for
   each of the print_*_begin_tag functions.

   print_*_empty_tag functions print simple tree nodes whose XML elements
   are empty (have nothing nested inside them, and use the special <.../>
   syntax.  These elements may have attributes in the tags.

   All the print_*_*_tag functions are called from xml_output_* functions
   to do the actual text output.  They are responsible for printing the
   indentation spaces, and for printing newlines after each tag.
   print_*_*_tag functions should never call each other.  If you need to
   do this, create an xml_output_* function to do all the print_*_*_tag
   calls instead.
*/

#include "config.h"
#include "system.h"
#include "sys/stat.h"

#include "tree.h"
#include "cp-tree.h"
#include "decl.h"
#include "rtl.h"
#include "varray.h"

/* Return non-zero if the given _DECL is an internally generated decl.  */
#define DECL_INTERNAL_P(d) (DECL_SOURCE_LINE(d) == 0)
/*#define DECL_INTERNAL_P(d) (strcmp (DECL_SOURCE_FILE (d), "<internal>")==0)*/

void do_xml_output                            PARAMS ((const char*));
    
static void xml_output_headers                PARAMS ((FILE*));
static void xml_output_namespace_decl         PARAMS ((FILE*, unsigned long, tree));
static void xml_output_type_decl              PARAMS ((FILE*, unsigned long, tree));
static void xml_output_record_type            PARAMS ((FILE*, unsigned long, tree));
static void xml_output_function_decl          PARAMS ((FILE*, unsigned long, tree));
static void xml_output_overload               PARAMS ((FILE*, unsigned long, tree));
static void xml_output_var_decl               PARAMS ((FILE*, unsigned long, tree));
static void xml_output_field_decl             PARAMS ((FILE*, unsigned long, tree));
static void xml_output_const_decl             PARAMS ((FILE*, unsigned long, tree));
static void xml_output_template_decl          PARAMS ((FILE*, unsigned long, tree));
static void xml_output_template_info          PARAMS ((FILE*, unsigned long, tree));
static void xml_output_typedef                PARAMS ((FILE*, unsigned long, tree));
static void xml_output_type                   PARAMS ((FILE*, unsigned long, tree));
static void xml_output_named_type             PARAMS ((FILE*, unsigned long, tree));
static void xml_output_void_type              PARAMS ((FILE*, unsigned long, tree));
static void xml_output_function_type          PARAMS ((FILE*, unsigned long, tree));
static void xml_output_method_type            PARAMS ((FILE*, unsigned long, tree));
static void xml_output_pointer_type           PARAMS ((FILE*, unsigned long, tree));
static void xml_output_reference_type         PARAMS ((FILE*, unsigned long, tree));
static void xml_output_offset_type            PARAMS ((FILE*, unsigned long, tree));
static void xml_output_array_type             PARAMS ((FILE*, unsigned long, tree));
static void xml_output_enumeral_type          PARAMS ((FILE*, unsigned long, tree));
static void xml_output_base_class             PARAMS ((FILE*, unsigned long, tree));
static void xml_output_base_type              PARAMS ((FILE*, unsigned long, tree));
static void xml_output_argument               PARAMS ((FILE*, unsigned long, tree, tree));
static void xml_output_returns                PARAMS ((FILE*, unsigned long, tree));
static void xml_output_template_argument      PARAMS ((FILE*, unsigned long, tree));
static void xml_output_qualified_name         PARAMS ((FILE*, unsigned long, tree));

static const char* xml_get_encoded_string             PARAMS ((tree));
static const char* xml_get_encoded_string_from_string PARAMS ((const char*));

static void print_indent                   PARAMS ((FILE*, unsigned long));
static tree reverse_opname_lookup          PARAMS ((tree));

/* Number of spaces to indent each nested element in XML output.  */
#define XML_NESTED_INDENT 2



/* Main XML output function.  Called by parser at the end of a translation
   unit.  Walk the entire translation unit starting at the global
   namespace.  Output all declarations.  */
void
do_xml_output (const char* filename)
{
  FILE *file;

  file = fopen (filename, "w");
  if (!file)
    cp_error ("could not open xml-dump file `%s'", filename);
  else
    {
    xml_output_headers (file);
    xml_output_namespace_decl (file, 0, global_namespace);
    fclose (file);
    }
}



/* Print XML begin tag for a namespace element.  */
static void
print_namespace_begin_tag (FILE* file, unsigned long indent, tree ns)
{
  print_indent (file, indent);
  if(ns == global_namespace)
    {
    fprintf (file, "<GlobalNamespace>\n");
    }
  else
    {
    const char* name = xml_get_encoded_string (DECL_NAME (ns));

    fprintf (file, "<Namespace name=\"%s\">\n", name);
    }
}


/* Print XML end tag for a namespace element.  */
static void
print_namespace_end_tag (FILE* file, unsigned long indent, tree ns)
{
  print_indent (file, indent);
  if(ns == global_namespace)
    {
    fprintf (file, "</GlobalNamespace>\n");
    }
  else
    {
    fprintf (file, "</Namespace>\n");
    }
}


/* Print XML begin tag for a namespace alias element.  */
static void
print_namespace_alias_begin_tag (FILE* file, unsigned long indent, tree ns)
{
  const char* name = xml_get_encoded_string (DECL_NAME (ns));

  print_indent (file, indent);
  fprintf (file,
           "<NamespaceAlias name=\"%s\">\n",
           name);
}


/* Print XML end tag for a namespace alias element.  */
static void
print_namespace_alias_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</NamespaceAlias>\n");
}


/* Print XML begin tag for a typedef element.  */
static void
print_typedef_begin_tag (FILE* file, unsigned long indent, tree td)
{
  const char* name = xml_get_encoded_string (DECL_NAME (td));

  print_indent (file, indent);
  fprintf (file,
           "<Typedef name=\"%s\">\n",
           name);
}


/* Print XML end tag for a typedef element.  */
static void
print_typedef_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Typedef>\n");  
}


/* Print XML begin tag for a class, struct, or union element.  */
static void
print_class_begin_tag (FILE* file, unsigned long indent, tree rt)
{
  const char* name = xml_get_encoded_string (DECL_NAME (TYPE_NAME (rt)));  
  const char* access = "";
  const char* mangled_name = mangle_type_string (rt);
  int is_abstract = (CLASSTYPE_PURE_VIRTUALS (rt) != 0);

  if (TREE_PRIVATE (TYPE_NAME (rt)))        access = "private";
  else if (TREE_PROTECTED (TYPE_NAME (rt))) access = "protected";
  else                          access = "public";  

  print_indent (file, indent);
  if (TREE_CODE(rt) == RECORD_TYPE)
    {
    if (CLASSTYPE_DECLARED_CLASS (rt))
      {
      fprintf (file,
               "<Class name=\"%s\" access=\"%s\" abstract=\"%d\" mangled_name=\"%s\">\n",
               name, access, is_abstract, mangled_name);
      }
    else
      {
      fprintf (file,
               "<Struct name=\"%s\" access=\"%s\" abstract=\"%d\" mangled_name=\"%s\">\n",
               name, access, is_abstract, mangled_name);
      }
    }
  else
    {
      fprintf (file,
               "<Union name=\"%s\" access=\"%s\" abstract=\"%d\" mangled_name=\"%s\">\n",
               name, access, is_abstract, mangled_name);
    }
}


/* Print XML end tag for a class, struct, or union element.  */
static void
print_class_end_tag (FILE* file, unsigned long indent, tree rt)
{
  print_indent (file, indent);
  if (TREE_CODE(rt) == RECORD_TYPE)
    {
    if (CLASSTYPE_DECLARED_CLASS (rt))
      {
      fprintf (file,
               "</Class>\n");
      }
    else
      {
      fprintf (file,
               "</Struct>\n");
      }
    }
  else
    {
      fprintf (file,
               "</Union>\n");
    }
}


/* Print XML begin tag for a constructor element.  */
static void
print_constructor_begin_tag (FILE* file, unsigned long indent, tree fd)
{
  const char* access = "";

  if (TREE_PRIVATE (fd))        access = "private";
  else if (TREE_PROTECTED (fd)) access = "protected";
  else                          access = "public";  

  print_indent (file, indent);
  fprintf (file,
           "<Constructor access=\"%s\">\n",
           access);
}


/* Print XML end tag for a constructor element.  */
static void
print_constructor_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Constructor>\n");
}


/* Print XML begin tag for a destructor element.  */
static void
print_destructor_begin_tag (FILE* file, unsigned long indent, tree fd)
{
  const char* access = "";
  int is_virtual;
  int is_pure_virtual;

  /* xml_output_record_type may call this with fd=0 if a trivial
     destructor is encountered.  In this case, set the values
     explicitly.  */
  if (fd)
    {
    if (TREE_PRIVATE (fd))        access = "private";
    else if (TREE_PROTECTED (fd)) access = "protected";
    else                          access = "public";
    
    if (DECL_VIRTUAL_P (fd))
      is_virtual = 1;
    else
      is_virtual = 0;
    
    if (DECL_PURE_VIRTUAL_P (fd))
      is_pure_virtual = 1;
    else
      is_pure_virtual = 0;
    }
  else
    {
    access = "public";
    is_virtual = 0;
    is_pure_virtual = 0;
    }

  print_indent (file, indent);
  fprintf (file,
           "<Destructor access=\"%s\" virtual=\"%d\" pure_virtual=\"%d\">\n",
           access, is_virtual, is_pure_virtual);
}


/* Print XML end tag for a destructor element.  */
static void
print_destructor_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Destructor>\n");
}


/* Print XML begin tag for a converter (type-conversion operator)
   element.  */
static void
print_converter_begin_tag (FILE* file, unsigned long indent, tree fd)
{
  const char* access = "";
  int is_const;
  int is_virtual;
  int is_pure_virtual;

  if (TREE_PRIVATE (fd))        access = "private";
  else if (TREE_PROTECTED (fd)) access = "protected";
  else                          access = "public";  

  if (DECL_CONST_MEMFUNC_P (fd))
    is_const = 1;
  else
    is_const = 0;

  if (DECL_VIRTUAL_P (fd))
    is_virtual = 1;
  else
    is_virtual = 0;

  if (DECL_PURE_VIRTUAL_P (fd))
    is_pure_virtual = 1;
  else
    is_pure_virtual = 0;
  
  print_indent (file, indent);
  fprintf (file,
           "<Converter access=\"%s\" const=\"%d\" virtual=\"%d\" pure_virtual=\"%d\">\n",
           access, is_const, is_virtual, is_pure_virtual);
}


/* Print XML end tag for a converter (type-conversion operator) element.  */
static void
print_converter_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Converter>\n");
}


/* Print XML begin tag for an operator function element.  */
static void
print_operator_function_begin_tag (FILE* file, unsigned long indent, tree fd)
{
  const char* name = xml_get_encoded_string (reverse_opname_lookup (DECL_NAME (fd)) );

  print_indent (file, indent);
  fprintf (file,
           "<OperatorFunction name=\"%s\">\n",
           name);
}


/* Print XML end tag for an operator function element.  */
static void
print_operator_function_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</OperatorFunction>\n");
}


/* Print XML begin tag for an operator method element.  */
static void
print_operator_method_begin_tag (FILE* file, unsigned long indent, tree fd)
{
  const char* name = xml_get_encoded_string (reverse_opname_lookup (DECL_NAME (fd)) );
  const char* access = "";
  int is_const;
  int is_virtual;
  int is_pure_virtual;
  int is_static;

  if (TREE_PRIVATE (fd))        access = "private";
  else if (TREE_PROTECTED (fd)) access = "protected";
  else                          access = "public";  

  if (DECL_CONST_MEMFUNC_P (fd))
    is_const = 1;
  else
    is_const = 0;

  if (DECL_VIRTUAL_P (fd))
    is_virtual = 1;
  else
    is_virtual = 0;

  if (DECL_PURE_VIRTUAL_P (fd))
    is_pure_virtual = 1;
  else
    is_pure_virtual = 0;

  if (DECL_NONSTATIC_MEMBER_FUNCTION_P (fd))
    is_static = 0;
  else
    is_static = 1;

  print_indent (file, indent);
  fprintf (file,
           "<OperatorMethod name=\"%s\" access=\"%s\" const=\"%d\" virtual=\"%d\" pure_virtual=\"%d\" static=\"%d\">\n",
           name, access, is_const, is_virtual, is_pure_virtual, is_static);
}


/* Print XML end tag for an operator method element.  */
static void
print_operator_method_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</OperatorMethod>\n");
}


/* Print XML begin tag for a method element.  */
static void
print_method_begin_tag (FILE* file, unsigned long indent, tree fd)
{
  const char* name = xml_get_encoded_string (DECL_NAME (fd));
  const char* access = "";
  int is_static;
  int is_const;
  int is_virtual;
  int is_pure_virtual;

  if (TREE_PRIVATE (fd))        access = "private";
  else if (TREE_PROTECTED (fd)) access = "protected";
  else                          access = "public";  

  if (DECL_NONSTATIC_MEMBER_FUNCTION_P (fd))
    is_static = 0;
  else
    is_static = 1;

  if (DECL_CONST_MEMFUNC_P (fd))
    is_const = 1;
  else
    is_const = 0;

  if (DECL_VIRTUAL_P (fd))
    is_virtual = 1;
  else
    is_virtual = 0;

  if (DECL_PURE_VIRTUAL_P (fd))
    is_pure_virtual = 1;
  else
    is_pure_virtual = 0;

  print_indent (file, indent);
  fprintf (file,
           "<Method name=\"%s\" access=\"%s\" static=\"%d\" const=\"%d\" virtual=\"%d\" pure_virtual=\"%d\">\n",
           name, access, is_static, is_const, is_virtual, is_pure_virtual);
}


/* Print XML end tag for a method element.  */
static void
print_method_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Method>\n");
}


/* Print XML begin tag for a function element.  */
static void
print_function_begin_tag (FILE* file, unsigned long indent, tree fd)
{
  const char* name = xml_get_encoded_string (DECL_NAME (fd));

  print_indent (file, indent);
  fprintf (file,
           "<Function name=\"%s\">\n",
           name);
}


/* Print XML end tag for a function element.  */
static void
print_function_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Function>\n");
}


/* Print XML begin tag for a function argument element.  */
static void
print_argument_begin_tag (FILE* file, unsigned long indent, tree arg)
{
  const char* name = "";

  if(arg && DECL_NAME (arg))
    name = xml_get_encoded_string (DECL_NAME (arg));
  
  print_indent (file, indent);
  fprintf (file,
           "<Argument name=\"%s\">\n",
           name);  
}


/* Print XML end tag for a function argument element.  */
static void
print_argument_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Argument>\n");
}


/* Print XML begin tag for a function return type element.  */
static void
print_returns_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<Returns>\n");
}


/* Print XML end tag for a function return type element.  */
static void
print_returns_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Returns>\n");
}


/* Print XML empty tag for a default argument element.  */
static void
print_default_argument_empty_tag (FILE* file, unsigned long indent, tree t)
{
  const char* value;

  value = xml_get_encoded_string_from_string (
    expr_as_string (t, 0));
  
  print_indent (file, indent);
  fprintf (file,
           "<DefaultArgument value=\"%s\"/>\n",
           value);
}


/* Print XML empty tag for an ellipsis at the end of an argument list.  */
static void
print_ellipsis_empty_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<Ellipsis/>\n");
}


/* Print XML begin tag for a variable declaration element.  */
static void
print_variable_begin_tag (FILE* file, unsigned long indent, tree vd)
{
  const char* name = xml_get_encoded_string (DECL_NAME (vd));
  
  print_indent (file, indent);
  fprintf (file,
           "<Variable name=\"%s\">\n",
           name);
}


/* Print XML end tag for a variable declaration element.  */
static void
print_variable_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Variable>\n");
}


/* Print XML empty tag for a variable initializer element.  */
static void
print_initializer_empty_tag (FILE* file, unsigned long indent, tree t)
{
  const char* value;
  
  if (!t || (t == error_mark_node)) return;
  
  value = xml_get_encoded_string_from_string (expr_as_string (t, 0));
  print_indent (file, indent);
  fprintf (file,
           "<Initializer value=\"%s\"/>\n",
           value);
}


/* Print XML begin tag for a field declaration element.  */
static void
print_field_begin_tag (FILE* file, unsigned long indent, tree vd)
{
  const char* name = xml_get_encoded_string (DECL_NAME (vd));
  
  print_indent (file, indent);
  fprintf (file,
           "<Field name=\"%s\">\n",
           name);
}


/* Print XML end tag for a field declaration element.  */
static void
print_field_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Field>\n");
}


/* Print XML begin tag for a const declaration element.  */
static void
print_enum_begin_tag (FILE* file, unsigned long indent, tree vd)
{
  const char* name = xml_get_encoded_string (DECL_NAME (vd));
  
  print_indent (file, indent);
  fprintf (file,
           "<Enum name=\"%s\">\n",
           name);
}


/* Print XML end tag for a const declaration element.  */
static void
print_enum_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Enum>\n");
}


/* Print XML begin tag for a type element.  */
static void
print_named_type_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);  
  fprintf (file,
           "<NamedType>\n");
}


/* Print XML end tag for a type element.  */
static void
print_named_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);  
  fprintf (file,
           "</NamedType>\n");
}


/* Print XML begin tag for a pointer type element.  */
static void
print_pointer_type_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<PointerType>\n");
}


/* Print XML end tag for a pointer type element.  */
static void
print_pointer_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file, "</PointerType>\n");
}


/* Print XML begin tag for a reference type element.  */
static void
print_reference_type_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<ReferenceType>\n");
}


/* Print XML end tag for a reference type element.  */
static void
print_reference_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file, "</ReferenceType>\n");
}


/* Print XML begin tag for a function type element.  */
static void
print_function_type_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<FunctionType>\n");
}


/* Print XML end tag for a function type element.  */
static void
print_function_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</FunctionType>\n");
}


/* Print XML begin tag for a method type element.  */
static void
print_method_type_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<MethodType>\n");
}


/* Print XML end tag for a method type element.  */
static void
print_method_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</MethodType>\n");
}


/* Print XML begin tag for a offset type element.  */
static void
print_offset_type_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<OffsetType>\n");
}


/* Print XML end tag for a offset type element.  */
static void
print_offset_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</OffsetType>\n");
}


/* Print XML begin tag for a array type element.  */
static void
print_array_type_begin_tag (FILE* file, unsigned long indent, tree at)
{
  const char* length = "";
  
  if (TYPE_DOMAIN (at))
    length = xml_get_encoded_string_from_string (
      expr_as_string (TYPE_MAX_VALUE (TYPE_DOMAIN (at)), 0));

  print_indent (file, indent);
  fprintf (file,
           "<ArrayType min=\"0\" max=\"%s\">\n",
           length);
}


/* Print XML end tag for a array type element.  */
static void
print_array_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</ArrayType>\n");
}


/* Print XML begin tag for an enumeration type element.  */
static void
print_enumeral_type_begin_tag (FILE* file, unsigned long indent, tree et)
{
  const char* name = xml_get_encoded_string (DECL_NAME (TYPE_NAME (et)));  
  const char* access = "";

  if (TREE_PRIVATE (TYPE_NAME (et)))        access = "private";
  else if (TREE_PROTECTED (TYPE_NAME (et))) access = "protected";
  else                          access = "public";  

  print_indent (file, indent);
  fprintf (file,
           "<EnumType name=\"%s\" access=\"%s\">\n",
           name, access);
}


/* Print XML end tag for an enumeration type element.  */
static void
print_enumeral_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</EnumType>\n");
}


/* Print XML begin element tag for a base type element.  */
static void
print_base_type_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<BaseType>\n");
}


/* Print XML end element tag for a base type element.  */
static void
print_base_type_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</BaseType>\n");
}


/* Print XML begin element tag for a base class element.  */
static void
print_base_class_begin_tag (FILE* file, unsigned long indent, tree binfo)
{
  const char* access = "";

  if (TREE_VIA_PUBLIC (binfo))         access = "public";
  else if (TREE_VIA_PROTECTED (binfo)) access = "protected";
  else                                 access = "private";

  print_indent (file, indent);
  fprintf (file,
           "<BaseClass access=\"%s\">\n",
           access);
}


/* Print XML end element tag for a base class element.  */
static void
print_base_class_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</BaseClass>\n");
}


/* Print XML begin tag for a template instantiation element.  */
static void
print_instantiation_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<Instantiation>\n");
}


/* Print XML end tag for a template instantiation element.  */
static void
print_instantiation_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</Instantiation>\n");
}


/* Print XML begin tag for a template argument element.  */
static void
print_template_argument_begin_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<TemplateArgument>\n");
}


/* Print XML end tag for a template argument element.  */
static void
print_template_argument_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</TemplateArgument>\n");
}


/* Print XML empty tag indicating that the current declaration is
   external.  */
static void
print_external_empty_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<External/>\n");
}


/* Print XML empty tag indicating that the current type is incomplete.  */
static void
print_incomplete_type_empty_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "<IncompleteType/>\n");
}


/* Print XML empty tag describing the location of a declaration in the
   source.  */
static void
print_location_empty_tag (FILE* file, unsigned long indent, tree d)
{  
  const char* source_file = DECL_SOURCE_FILE (d);
  unsigned int source_line = DECL_SOURCE_LINE (d);
  
  print_indent (file, indent);
  fprintf (file,
           "<Location file=\"%s\" line=\"%d\"/>\n",
           source_file, source_line);
}


/* Print XML empty tag describing the cv-qualifiers of a type.  */
static void
print_cv_qualifiers_empty_tag (FILE* file, unsigned long indent, tree t)
{  
  int is_const = CP_TYPE_CONST_P (t);
  int is_volatile = CP_TYPE_VOLATILE_P (t);
  int is_restrict = CP_TYPE_RESTRICT_P (t);  
  
  /* If there are no qualifiers, don't bother with the output.  */
  if (!is_const && !is_volatile && !is_restrict) return;
  
  print_indent (file, indent);
  fprintf (file,
           "<CvQualifiers const=\"%d\" volatile=\"%d\" restrict=\"%d\"/>\n",
           is_const, is_volatile, is_restrict);
}


/* Print XML begin tag for a name qualifier element.  */
static void
print_name_qualifier_begin_tag (FILE* file, unsigned long indent, tree t)
{
  const char* name = xml_get_encoded_string(t);
  print_indent (file, indent);
  fprintf (file,
           "<NameQualifier name=\"%s\">\n",
           name);
}


/* Print XML end tag for a name qualifier element.  */
static void
print_name_qualifier_end_tag (FILE* file, unsigned long indent)
{
  print_indent (file, indent);
  fprintf (file,
           "</NameQualifier>\n");
}


/* Print XML empty tag for a qualified name element.  */
static void
print_qualified_name_empty_tag (FILE* file, unsigned long indent, tree t)
{
  const char* name = xml_get_encoded_string(t);
  print_indent (file, indent);
  fprintf (file,
           "<QualifiedName name=\"%s\"/>\n",
           name);
}


/* Print XML empty tag for a scope reference element (SCOPE_REF).  */
static void
print_scope_ref_empty_tag (FILE* file, unsigned long indent, tree t)
{
  const char* class_name =
    xml_get_encoded_string (DECL_NAME (TYPE_NAME (TREE_OPERAND (t, 0))));
  const char* field_name =
    xml_get_encoded_string (TREE_OPERAND (t, 1));
  print_indent (file, indent);
  fprintf (file,
           "<ScopeRef class=\"%s\" field=\"%s\"/>\n",
           class_name, field_name);
}


/* Print XML empty tag for an integer literal element (INTEGER_CST).  */
static void
print_integer_cst_empty_tag (FILE * file, unsigned long indent, tree t)
{
  print_indent (file, indent);
  fprintf (file, "<Integer value=\"");
  if (TREE_INT_CST_HIGH (t) == 0)
    {
    fprintf (file, HOST_WIDE_INT_PRINT_UNSIGNED, TREE_INT_CST_LOW (t));
    }
  else if ((TREE_INT_CST_HIGH (t) == -1) && (TREE_INT_CST_LOW (t) != 0))
    {
    fprintf (file, "-");
    fprintf (file, HOST_WIDE_INT_PRINT_UNSIGNED, -TREE_INT_CST_LOW (t));
    }
  else
    {
    fprintf (file, HOST_WIDE_INT_PRINT_DOUBLE_HEX,
             TREE_INT_CST_HIGH (t), TREE_INT_CST_LOW (t));
    }
  fprintf(file, "\"/>\n");
}


/* Print XML empty tag naming an enumeration value.  */
static void
print_enum_value_empty_tag (FILE* file, unsigned long indent, tree v)
{  
  const char* value =  xml_get_encoded_string (v);
  
  print_indent (file, indent);
  fprintf (file,
           "<EnumValue name=\"%s\"/>\n",
           value);
}


/* Print XML empty tag describing an unimplemented TREE_CODE that has been
   encountered.  */
static void
print_unimplemented_empty_tag (FILE* file, unsigned long indent, tree t,
                               const char* func)
{
  int tree_code = TREE_CODE (t);
  
  print_indent (file, indent);
  fprintf (file,
           "<Unimplemented tree_code=\"%d\" tree_code_name=\"%s\""
           " xml_c_location=\"%s\"/>\n",
           tree_code, tree_code_name [tree_code], func);
}


/* Output the XML file's headers.  */
void
xml_output_headers (FILE* file)
{
  fprintf (file,
           "<?xml version=\"1.0\"?>\n");/* encoding="ISO-8859-1"?> */
}


/* Output a NAMESPACE_DECL.  Prints beginning and ending tags, and all
   the namespace's member declarations in between.  */
void
xml_output_namespace_decl (FILE* file, unsigned long indent, tree ns)
{
  if(ns == fake_std_node) return;
    
  /* Only walk a real namespace.  */
  if (!DECL_NAMESPACE_ALIAS (ns))
    {
    tree cur_decl;

    print_namespace_begin_tag(file, indent, ns);

    for (cur_decl = cp_namespace_decls(ns); cur_decl;
         cur_decl = TREE_CHAIN (cur_decl))
      {
      switch (TREE_CODE (cur_decl))
        {
        case NAMESPACE_DECL:
          xml_output_namespace_decl(file, indent+XML_NESTED_INDENT, cur_decl);
          break;
        case TYPE_DECL:
          xml_output_type_decl(file, indent+XML_NESTED_INDENT, cur_decl);
          break;          
        case FUNCTION_DECL:
          xml_output_function_decl (file, indent+XML_NESTED_INDENT, cur_decl);
          break;
        case VAR_DECL:
          xml_output_var_decl (file, indent+XML_NESTED_INDENT, cur_decl);
          break;
        case TREE_LIST:
          xml_output_overload (file, indent+XML_NESTED_INDENT,
                               TREE_VALUE (cur_decl));
          break;
        case CONST_DECL:
          xml_output_const_decl (file, indent+XML_NESTED_INDENT, cur_decl);
          break;
        case TEMPLATE_DECL:
          xml_output_template_decl (file, indent+XML_NESTED_INDENT, cur_decl);
          break;
        case RESULT_DECL:
        case USING_DECL:
          /* case THUNK_DECL: */
          /* This is compiler-generated.  Just ignore it.  */
          break;
        default:
          print_unimplemented_empty_tag (file, indent+XML_NESTED_INDENT,
                                         cur_decl, "xml_output_namespace_decl");
        }
      }
    
    print_namespace_end_tag(file, indent, ns);
    }
  /* If it is a namespace alias, just indicate that.  */
  else
    {
    tree real_ns = ns;
    /* Find the real namespace.  */
    while (DECL_NAMESPACE_ALIAS (real_ns))
      {
      real_ns = DECL_NAMESPACE_ALIAS (real_ns);
      }

    print_namespace_alias_begin_tag (file, indent, ns);
    xml_output_qualified_name (file, indent+XML_NESTED_INDENT, real_ns);
    print_namespace_alias_end_tag (file, indent);
    }
}


/* When a class, struct, or union type is defined, it is automatically
   given a member typedef to itself.  Given a RECORD_TYPE or
   UNION_TYPE, this returns that field's name.  Although this should
   never happen, 0 is returned when the field cannot be found.  */
static tree
xml_find_self_typedef_name (tree rt)
{
  tree field;
  for (field = TYPE_FIELDS (rt) ; field ; field = TREE_CHAIN (field))
    {
    /* The field must be artificial because someone could have written
       their own typedef of the class's name.  */
    if ((TREE_CODE (field) == TYPE_DECL)
        && (TREE_TYPE (field) == rt)
        && DECL_ARTIFICIAL (field))
      {
      return DECL_NAME (field);
      }
    }

  return 0;
}


/* Dispatch output of a TYPE_DECL.  This is either a typedef, or a new
   type (class/struct/union) definition.  */
void
xml_output_type_decl (FILE* file, unsigned long indent, tree td)
{
  /* Get the type from the TYPE_DECL.  We don't want to use complete_type
     because it may modify something.  We are doing a read-only dump.  */
  tree t = TREE_TYPE (td);

  /* Don't process any internally generated declarations.  */
  if (DECL_INTERNAL_P (td)) return;

  switch (TREE_CODE (t))
    {
    case ARRAY_TYPE:
    case LANG_TYPE:
    case POINTER_TYPE:
    case REFERENCE_TYPE:
    case INTEGER_TYPE:
    case VOID_TYPE:
    case BOOLEAN_TYPE:
    case REAL_TYPE:
    case COMPLEX_TYPE:
    case FUNCTION_TYPE:
    case TYPENAME_TYPE:
    case TEMPLATE_TYPE_PARM:
      /* A typedef to a predefined type.  */
      xml_output_typedef (file, indent, td);
      break;
    case ENUMERAL_TYPE:
      if (DECL_ORIGINAL_TYPE (td))
        {
        /* A typedef to an existing enumeral type.  */
        xml_output_typedef (file, indent, td);
        }
      else
        {
        /* This is the declaration of a new enumeration type.  */
        xml_output_enumeral_type (file, indent, t);
        }
      break;
    case UNION_TYPE:
    case RECORD_TYPE:
      if ((TREE_CODE (t) == RECORD_TYPE) && TYPE_PTRMEMFUNC_P (t))
        {
        /* A typedef to a pointer to member.  */
        xml_output_typedef (file, indent, td);
        }
      /* This is a complicated test to tell apart a real class declaration
         or definition from a typedef to it.  This must be able to detect
         the difference for the following cases:

            Code:                           Meaning:
         1.  struct A {};                    "struct A"
         2.  struct A {}; typedef A B;       "struct A" "typedef A B"
         3.  typedef struct {} A;            "struct A"
         4.  typedef struct A {} A;          "struct A"
         5.  typedef struct A {} B;          "struct A" "typedef A B"

         DECL_IMPLICIT_TYPEDEF_P will recognize type TYPE_DECL that
         points to the real class definition for cases 1, 2, and 5,
         and is sufficient to make the decision.

         For cases 3 and 4, the second test makes sure the context of
         the TYPE_DECL matches the context of the RECORD_TYPE, which
         can only happen for a TYPE_DECL in the same scope as the
         class.  It then looks at the RECORD_TYPE to find its
         artificial member that is a typedef to its own class.  The
         name of this field is compared to the name of the original
         TYPE_DECL.  If it matches, then it is assumed that the
         TYPE_DECL corresponds to the class's definition.  Otherwise,
         it is assumed that it is simply a typedef.  */
      else if (DECL_IMPLICIT_TYPEDEF_P (td)
               || (((DECL_CONTEXT (td) == TYPE_CONTEXT (t))
                    && (xml_find_self_typedef_name (t) == DECL_NAME (td)))))
        {
        /* A new type definition.  */
        xml_output_record_type (file, indent, t);
        }
      else
        {
        /* A typedef to an existing class, struct, or union type.  */
        xml_output_typedef (file, indent, td);
        }
      break;
    default:
      print_unimplemented_empty_tag (file, indent, t,
                                     "xml_output_type_decl");
    }
}


/* Output a RECORD_TYPE that is not a pointer-to-member-function.  Prints
   beginning and ending tags, and all class member declarations between.
   Also handles a UNION_TYPE.  */
void
xml_output_record_type (FILE* file, unsigned long indent, tree rt)
{
  tree field;
  tree func;
  int destructorFound = 0;

  print_class_begin_tag (file, indent, rt);
  print_location_empty_tag (file, indent+XML_NESTED_INDENT, TYPE_NAME (rt));

  if (!COMPLETE_TYPE_P (rt))
    {
    print_incomplete_type_empty_tag (file, indent+XML_NESTED_INDENT);
    }
  
  if (CLASSTYPE_TEMPLATE_INFO (rt))
    {
    xml_output_template_info (file, indent+XML_NESTED_INDENT,
                              CLASSTYPE_TEMPLATE_INFO (rt));
    }  
  
  /* Output all the non-method declarations in the class.  */
  for (field = TYPE_FIELDS (rt) ; field ; field = TREE_CHAIN (field))
    {
    switch(TREE_CODE(field))
      {
      case TYPE_DECL:
        if (TREE_TYPE (field) == rt)
          {
          /* A class or struct internally typedefs itself.  */
          xml_output_typedef (file, indent+XML_NESTED_INDENT, field);
          }
        else
          {
          /* A nested type declaration.  */
          xml_output_type_decl (file, indent+XML_NESTED_INDENT, field);
          }
        break;          
      case FIELD_DECL:
        xml_output_field_decl (file, indent+XML_NESTED_INDENT, field);
        break;          
      case VAR_DECL:
        xml_output_var_decl (file, indent+XML_NESTED_INDENT, field);
        break;
      case CONST_DECL:
        xml_output_const_decl (file, indent+XML_NESTED_INDENT, field);
        break;
      case TEMPLATE_DECL:
        xml_output_template_decl (file, indent+XML_NESTED_INDENT, field);
        break;
      case USING_DECL:
        /* Ignore the using decl.  */
	break;
      case RESULT_DECL:
      default:
        print_unimplemented_empty_tag (file, indent+XML_NESTED_INDENT, field,
                                       "xml_output_record_type fields");
      }
    }

  /* Output all the method declarations in the class.  */
  for (func = TYPE_METHODS (rt) ; func ; func = TREE_CHAIN (func))
    {
    switch (TREE_CODE (func))
      {
      case FUNCTION_DECL:
        if (DECL_DESTRUCTOR_P (func)) 
          destructorFound = 1;
        xml_output_function_decl (file, indent+XML_NESTED_INDENT, func);
        break;
      case TEMPLATE_DECL:
        xml_output_template_decl (file, indent+XML_NESTED_INDENT, func);
        break;
      default:
        print_unimplemented_empty_tag (file, indent+XML_NESTED_INDENT, func,
                                       "xml_output_record_type methods");
      }
    }

  /* If no destructor was encountered, then the class has a trivial destructor.
     For some reason, the implicit declaration was not there, so we catch
     the case here.  */
  if(!destructorFound && TYPE_HAS_TRIVIAL_DESTRUCTOR (rt))
    {
    print_destructor_begin_tag (file, indent+XML_NESTED_INDENT, 0);
    print_destructor_end_tag (file, indent+XML_NESTED_INDENT);
    }
  
  /* Output all the base classes.  */
  {
  tree binfo = TYPE_BINFO (rt);
  tree binfos = BINFO_BASETYPES (binfo);
  int n_baselinks = binfos? TREE_VEC_LENGTH (binfos) : 0;
  int i;

  for (i = 0; i < n_baselinks; i++)
    {
    tree base_binfo = TREE_VEC_ELT (binfos, i);
    if (base_binfo)
      {
      /* Output this base class.  */
      xml_output_base_class (file, indent+XML_NESTED_INDENT, base_binfo);
      }
    }
  }
  
  print_class_end_tag (file, indent, rt);
}


/* Output for a FUNCTION_DECL.  Prints beginning and ending tags, and
   the argument list between them.  */
void
xml_output_function_decl (FILE* file, unsigned long indent, tree fd)
{
  void (*end_tag_print)(FILE*, unsigned long) = NULL;
  tree arg;
  tree arg_type;

  /* Don't process any internally generated declarations.  */
  if (DECL_INTERNAL_P (fd)) return;

  /* Don't process any compiler-generated functions except constructors
     and destructors.  */
  if (DECL_ARTIFICIAL(fd)
      && !DECL_CONSTRUCTOR_P (fd)
      && !DECL_DESTRUCTOR_P (fd)) return;

  /* Don't output the cloned functions.  */
  if (DECL_CLONED_FUNCTION_P (fd)) return;

  /* Print out the begin tag for this type of function.  */
  if (DECL_CONSTRUCTOR_P (fd))
    {
    /* A class constructor.  */
    print_constructor_begin_tag (file, indent, fd);
    end_tag_print = print_constructor_end_tag;
    }
  else if (DECL_DESTRUCTOR_P (fd))
    {
    /* A class destructor.  */
    print_destructor_begin_tag (file, indent, fd);

    /* Don't print any arguments for a destructor.  */
    print_destructor_end_tag (file, indent);
    return;
    }
  else if (DECL_OVERLOADED_OPERATOR_P (fd))
    {
    if (DECL_CONV_FN_P (fd))
      {
      /* A type-conversion operator in a class.  */
      print_converter_begin_tag (file, indent, fd);
      end_tag_print = print_converter_end_tag;
      }
    else
      {
      if (DECL_FUNCTION_MEMBER_P (fd))
        {
        /* An operator in a class.  */
        print_operator_method_begin_tag (file, indent, fd);
        end_tag_print = print_operator_method_end_tag;
        }
      else
        {
        /* An operator in a namespace.  */
        print_operator_function_begin_tag (file, indent, fd);
        end_tag_print = print_operator_function_end_tag;
        }
      }
    }
  else
    {
    if (DECL_FUNCTION_MEMBER_P (fd))
      {
      /* A member of a class.  */
      print_method_begin_tag (file, indent, fd);
      end_tag_print = print_method_end_tag;
      }
    else
      {
      /* A member of a namespace.  */
      print_function_begin_tag (file, indent, fd);
      end_tag_print = print_function_end_tag;
      }
    }

  /* Print the location information for this function.  */
  print_location_empty_tag (file, indent+XML_NESTED_INDENT, fd);

  if (DECL_TEMPLATE_INFO (fd))
    {
    xml_output_template_info (file, indent+XML_NESTED_INDENT,
                              DECL_TEMPLATE_INFO (fd));
    }

  if (DECL_REALLY_EXTERN (fd))
    {
    print_external_empty_tag (file, indent+XML_NESTED_INDENT);
    }

  if (!DECL_CONSTRUCTOR_P (fd) && !DECL_DESTRUCTOR_P (fd))
    {
    /* Print out the return type information for this function.  */
    xml_output_returns (file, indent+XML_NESTED_INDENT,
                        TREE_TYPE (TREE_TYPE (fd)));
    }
  
  /* Prepare to iterator through argument list.  */
  arg = DECL_ARGUMENTS (fd);
  arg_type = TYPE_ARG_TYPES (TREE_TYPE (fd));
  if (DECL_NONSTATIC_MEMBER_FUNCTION_P (fd))
    {
    /* Skip "this" argument.  */
    if(arg) arg = TREE_CHAIN (arg);
    arg_type = TREE_CHAIN (arg_type);
    }  
  
  /* Print out the argument list for this function.  */
  while (arg_type && (arg_type != void_list_node))
    {
    xml_output_argument (file, indent+XML_NESTED_INDENT, arg, arg_type);
    if(arg) arg = TREE_CHAIN (arg);
    arg_type = TREE_CHAIN (arg_type);
    }

  if(!arg_type)
    {
    /* Function has variable number of arguments.  */
    print_ellipsis_empty_tag(file, indent+XML_NESTED_INDENT);
    }

  /* Print out the end tag for this type of function.  */
  if(end_tag_print)
    end_tag_print (file, indent);
}


/* Output for an OVERLOAD.  Output all the functions in the overload list.   */
void
xml_output_overload (FILE* file, unsigned long indent, tree o)
{
  tree cur_overload = o;
  while (cur_overload)
    {
    xml_output_function_decl (file, indent, OVL_CURRENT (cur_overload));
    cur_overload = OVL_NEXT (cur_overload);
    }
}


/* Output for a VAR_DECL.  The name and type of the variable are output,
   as well as the initializer if it exists.  */
void
xml_output_var_decl (FILE* file, unsigned long indent, tree vd)
{
  /* Don't process any internally generated declarations.  */
  if (DECL_INTERNAL_P (vd)) return;

  print_variable_begin_tag (file, indent, vd);

  print_location_empty_tag (file, indent+XML_NESTED_INDENT, vd);

  if (DECL_EXTERNAL (vd))
    {
    print_external_empty_tag (file, indent+XML_NESTED_INDENT);
    }

  xml_output_type (file, indent+XML_NESTED_INDENT, TREE_TYPE (vd));

  print_initializer_empty_tag (file, indent+XML_NESTED_INDENT, DECL_INITIAL (vd));

  print_variable_end_tag (file, indent);
}


/* Output for a FIELD_DECL.  The name and type of the variable are output,
   as well as the initializer if it exists.  */
void
xml_output_field_decl (FILE* file, unsigned long indent, tree fd)
{
  /* Don't process any internally generated declarations.  */
  if (DECL_INTERNAL_P (fd)) return;

  print_field_begin_tag (file, indent, fd);
  print_location_empty_tag (file, indent+XML_NESTED_INDENT, fd);
  /* TODO: handle bit field case.  */
  xml_output_type (file, indent+XML_NESTED_INDENT, TREE_TYPE (fd));
  print_field_end_tag (file, indent);
}


/* Output for a CONST_DECL.  The name and type of the constant are output,
   as well as the value.  */
void
xml_output_const_decl (FILE* file, unsigned long indent, tree cd)
{
  /* Don't process any internally generated declarations.  */
  if (DECL_INTERNAL_P (cd)) return;

  print_enum_begin_tag (file, indent, cd);
  print_location_empty_tag (file, indent+XML_NESTED_INDENT, cd);
  xml_output_type (file, indent+XML_NESTED_INDENT, TREE_TYPE (cd));
  print_initializer_empty_tag (file, indent+XML_NESTED_INDENT,
                          DECL_INITIAL (cd));
  print_enum_end_tag (file, indent);
}


/* Output for a TEMPLATE_DECL.  The set of specializations (including
   instantiations) is output.  */
void
xml_output_template_decl (FILE* file, unsigned long indent, tree td)
{
  tree tl;
  
  /* Don't process any internally generated declarations.  */
  if (DECL_INTERNAL_P (td)) return;

  for (tl = DECL_TEMPLATE_SPECIALIZATIONS (td);
       tl ; tl = TREE_CHAIN (tl))
    {
    tree ts = TREE_VALUE (tl);
    switch (TREE_CODE (ts))
      {
      case FUNCTION_DECL:
        xml_output_function_decl (file, indent, ts);
        break;
      case TEMPLATE_DECL:
        break;
      default:
        print_unimplemented_empty_tag (file, indent, ts,
                                       "xml_output_template_decl SPECIALIZATIONS");
      }
    }

  for (tl = DECL_TEMPLATE_INSTANTIATIONS (td);
       tl ; tl = TREE_CHAIN (tl))
    {
    tree ts = TYPE_NAME (TREE_VALUE (tl));
    switch (TREE_CODE (ts))
      {
      case TYPE_DECL:
        xml_output_type_decl (file, indent, ts);
        break;
      default:
        print_unimplemented_empty_tag (file, indent, ts,
                                       "xml_output_template_decl INSTANTIATIONS");
      }
    }
}


/* Output for TEMPLATE_INFO.  The set of template parameters used
   is output.  */
void
xml_output_template_info (FILE* file, unsigned long indent, tree ti)
{
  tree arg_vec;
  int num_args;
  int i;

  print_instantiation_begin_tag (file, indent);

  arg_vec = TI_ARGS (ti);
  num_args = TREE_VEC_LENGTH (arg_vec);
  for (i=0 ; i < num_args ; ++i)
    {
    tree arg = TREE_VEC_ELT (arg_vec, i);
    xml_output_template_argument (file, indent+XML_NESTED_INDENT, arg);
    }

  print_instantiation_end_tag (file, indent);  
}


/* Output for a typedef.  The name and associated type are output.  */
void
xml_output_typedef (FILE* file, unsigned long indent, tree td)
{
  /* If the typedef points to its own name in the same context, ignore
     it.  This can happen for code like "typedef struct {} A;".  */
  if(!DECL_ORIGINAL_TYPE (td)
     && TREE_TYPE (td) && TYPE_NAME (TREE_TYPE (td))
     && (DECL_NAME (td) == DECL_NAME (TYPE_NAME (TREE_TYPE (td))))
     && (DECL_CONTEXT (td) == DECL_CONTEXT (TYPE_NAME (TREE_TYPE (td)))))
     return;

  print_typedef_begin_tag (file, indent, td);
  print_location_empty_tag (file, indent+XML_NESTED_INDENT, td);
  
  /* Get the original type out of the typedef, if any.  */
  if (DECL_ORIGINAL_TYPE (td))
    xml_output_type (file, indent+XML_NESTED_INDENT,
                     DECL_ORIGINAL_TYPE (td));
  else
    xml_output_type (file, indent+XML_NESTED_INDENT,
                     TREE_TYPE (td));
  print_typedef_end_tag (file, indent);
}


/* Output for a *_TYPE.  This nests type definitions that contain other
   types (like pointers, references, functions, method, and arrays).  */
void
xml_output_type (FILE* file, unsigned long indent, tree t)
{
  switch (TREE_CODE (t))
    {
    case ARRAY_TYPE:
      xml_output_array_type (file, indent, t);
      break;
    case POINTER_TYPE:
      xml_output_pointer_type (file, indent, t);
      break;
    case REFERENCE_TYPE:
      xml_output_reference_type (file, indent, t);
      break;
    case FUNCTION_TYPE:
      xml_output_function_type (file, indent, t);
      break;
    case METHOD_TYPE:
      xml_output_method_type (file, indent, t);
      break;
    case OFFSET_TYPE:
      xml_output_offset_type (file, indent, t);
      break;
    case RECORD_TYPE:
      if (TYPE_PTRMEMFUNC_P (t))
        {
        /* Pointer-to-member-functions are stored ina RECORD_TYPE.  */
        xml_output_type (file, indent, TYPE_PTRMEMFUNC_FN_TYPE (t));
        }
      else
        {
        /* This is a struct or class type, just output its name.  */
        xml_output_named_type (file, indent, t);
        }
      break;
    case LANG_TYPE:
    case INTEGER_TYPE:
    case BOOLEAN_TYPE:
    case REAL_TYPE:
    case COMPLEX_TYPE:
    case ENUMERAL_TYPE:
    case UNION_TYPE:
    case TYPENAME_TYPE:
    case TEMPLATE_TYPE_PARM:
      xml_output_named_type (file, indent, t);
      break;
    case SCOPE_REF:
      print_scope_ref_empty_tag (file, indent, t);
      break;
    case TEMPLATE_PARM_INDEX:
      xml_output_type (file, indent, TREE_TYPE (t));
      break;
    case VOID_TYPE:
      xml_output_void_type (file, indent, t);
      break;
    case INTEGER_CST:
      print_integer_cst_empty_tag (file, indent, t);
      break;
    default:
      print_unimplemented_empty_tag (file, indent, t, "xml_output_type");
    }
}


/* Output for a normal named type.  */
void
xml_output_named_type (FILE* file, unsigned long indent, tree t)
{
  print_named_type_begin_tag (file, indent);
  print_cv_qualifiers_empty_tag (file, indent+XML_NESTED_INDENT, t);
  xml_output_qualified_name (file, indent+XML_NESTED_INDENT, t);
  print_named_type_end_tag (file, indent);
}


/* Output for a void type.  */
void
xml_output_void_type (FILE* file, unsigned long indent, tree t)
{
  print_named_type_begin_tag (file, indent);
  print_cv_qualifiers_empty_tag (file, indent+XML_NESTED_INDENT, t);
  xml_output_qualified_name (file, indent+XML_NESTED_INDENT,
                             TYPE_MAIN_VARIANT (t));
  print_named_type_end_tag (file, indent);
}


/* Output for a FUNCTION_TYPE.  */
void
xml_output_function_type (FILE* file, unsigned long indent, tree t)
{
  tree arg_type;
  
  print_function_type_begin_tag (file, indent);

  /* Print out the return type information for this function.  */
  xml_output_returns (file, indent+XML_NESTED_INDENT,
                      TREE_TYPE (t));
  
  /* Prepare to iterator through argument list.  */
  arg_type = TYPE_ARG_TYPES (t);

  /* Print out the argument list for this function.  */
  while (arg_type && (arg_type != void_list_node))
    {
    xml_output_argument (file, indent+XML_NESTED_INDENT, NULL, arg_type);
    arg_type = TREE_CHAIN (arg_type);
    }

  if(arg_type != void_list_node)
    {
    /* Function has variable number of arguments.  */
    print_ellipsis_empty_tag(file, indent+XML_NESTED_INDENT);
    }

  print_function_type_end_tag (file, indent);
}


/* Output for a METHOD_TYPE.  */
void
xml_output_method_type (FILE* file, unsigned long indent, tree t)
{
  tree arg_type;
  
  print_method_type_begin_tag (file, indent);

  xml_output_base_type (file, indent+XML_NESTED_INDENT,
                        TYPE_METHOD_BASETYPE (t));                         

  /* Print out the return type information for this method.  */
  xml_output_returns (file, indent+XML_NESTED_INDENT,
                      TREE_TYPE (t));
  
  /* Prepare to iterator through argument list.  */
  arg_type = TYPE_ARG_TYPES (t);

  /* Skip "this" argument.  */
  arg_type = TREE_CHAIN (arg_type);

  /* Print out the argument list for this method.  */
  while (arg_type && (arg_type != void_list_node))
    {
    xml_output_argument (file, indent+XML_NESTED_INDENT, NULL, arg_type);
    arg_type = TREE_CHAIN (arg_type);
    }

  if(arg_type != void_list_node)
    {
    /* Method has variable number of arguments.  */
    print_ellipsis_empty_tag(file, indent+XML_NESTED_INDENT);
    }

  print_cv_qualifiers_empty_tag (file, indent+XML_NESTED_INDENT, t);

  print_method_type_end_tag (file, indent);
}


/* Output for a POINTER_TYPE.  */
void
xml_output_pointer_type (FILE* file, unsigned long indent, tree t)
{
  print_pointer_type_begin_tag (file, indent);
  print_cv_qualifiers_empty_tag (file, indent+XML_NESTED_INDENT, t);
  xml_output_type (file, indent+XML_NESTED_INDENT, TREE_TYPE (t));
  print_pointer_type_end_tag (file, indent);
}


/* Output for a REFERENCE_TYPE.  */
void
xml_output_reference_type (FILE* file, unsigned long indent, tree t)
{
  print_reference_type_begin_tag (file, indent);
  print_cv_qualifiers_empty_tag (file, indent+XML_NESTED_INDENT, t);
  xml_output_type (file, indent+XML_NESTED_INDENT, TREE_TYPE (t));
  print_reference_type_end_tag (file, indent);
}


/* Output for an OFFSET_TYPE.  */
void
xml_output_offset_type (FILE* file, unsigned long indent, tree t)
{
  print_offset_type_begin_tag (file, indent);
  xml_output_base_type (file, indent+XML_NESTED_INDENT,
                        TYPE_OFFSET_BASETYPE (t));
  print_cv_qualifiers_empty_tag (file, indent+XML_NESTED_INDENT, t);
  xml_output_type (file, indent+XML_NESTED_INDENT, TREE_TYPE (t));
  print_offset_type_end_tag (file, indent);
}


/* Output for an ARRAY_TYPE.  */
void
xml_output_array_type (FILE* file, unsigned long indent, tree t)
{
  print_array_type_begin_tag (file, indent, t);
  print_cv_qualifiers_empty_tag (file, indent+XML_NESTED_INDENT, t);
  xml_output_type (file, indent+XML_NESTED_INDENT, TREE_TYPE (t));
  print_array_type_end_tag (file, indent);
}


/* Output for an ENUMERAL_TYPE.  */
void
xml_output_enumeral_type (FILE* file, unsigned long indent, tree t)
{
  tree tv;
  print_enumeral_type_begin_tag (file, indent, t);
  print_location_empty_tag (file, indent+XML_NESTED_INDENT, TYPE_NAME (t));

  /* Output the list of possible values for the enumeration type.  */
  for (tv = TYPE_VALUES (t); tv ; tv = TREE_CHAIN (tv))
    {
    print_enum_value_empty_tag(file, indent+XML_NESTED_INDENT,
                               TREE_PURPOSE (tv));
    }
  print_enumeral_type_end_tag (file, indent);
}


/* Output a base class for a RECORD_TYPE or UNION_TYPE.  */
void
xml_output_base_class (FILE* file, unsigned long indent, tree binfo)
{
  print_base_class_begin_tag (file, indent, binfo);
  xml_output_qualified_name (file, indent+XML_NESTED_INDENT,
                                   BINFO_TYPE (binfo));
  print_base_class_end_tag (file, indent);
}


/* Output a base type for a METHOD_TYPE or OFFSET_TYPE.  */
void
xml_output_base_type (FILE* file, unsigned long indent, tree t)
{
  print_base_type_begin_tag (file, indent);
  xml_output_qualified_name (file, indent+XML_NESTED_INDENT, t);
  print_base_type_end_tag (file, indent);
}


/* Output for a PARM_DECL / TREE_LIST corresponding to a function argument.  */
void
xml_output_argument (FILE* file, unsigned long indent, tree pd, tree tl)
{
  /* Don't process any compiler-generated arguments.  These occur for
     things like constructors of classes with virtual inheritance.  */
  if (pd && DECL_ARTIFICIAL (pd)) return;

  print_argument_begin_tag (file, indent, pd);

  if (pd && DECL_ARG_TYPE_AS_WRITTEN (pd))
    xml_output_type (file, indent+XML_NESTED_INDENT, DECL_ARG_TYPE_AS_WRITTEN (pd));
/*  else if (pd && DECL_ARG_TYPE (pd))
      xml_output_type (file, indent+XML_NESTED_INDENT, DECL_ARG_TYPE (pd)); */
  else if (pd && TREE_TYPE (pd))
    xml_output_type (file, indent+XML_NESTED_INDENT, TREE_TYPE (pd));
  else
    xml_output_type (file, indent+XML_NESTED_INDENT, TREE_VALUE (tl));
  
  if (TREE_PURPOSE (tl))
    print_default_argument_empty_tag (file, indent+XML_NESTED_INDENT,
                                      TREE_PURPOSE (tl));

  print_argument_end_tag (file, indent);
}


/* Output the return type of a function.  */
void
xml_output_returns (FILE* file, unsigned long indent, tree t)
{
  print_returns_begin_tag (file, indent);
  xml_output_type (file, indent+XML_NESTED_INDENT, t);
  print_returns_end_tag (file, indent);
}


/* Output the given template argument.  */
void
xml_output_template_argument (FILE* file, unsigned long indent, tree arg)
{
  print_template_argument_begin_tag (file, indent);
  xml_output_type (file, indent + XML_NESTED_INDENT, arg);
  print_template_argument_end_tag (file, indent);
}


/* Given a list of qualifiers, walk it recursively, and output it in
   nested form.  */
static void
output_qualifier_list (FILE* file, unsigned long indent, tree qualifiers)
{
  tree next = TREE_CHAIN (qualifiers);
  tree decl = TREE_VALUE (qualifiers);
  
  if (next)
    {
    print_name_qualifier_begin_tag (file, indent, DECL_NAME (decl));
    output_qualifier_list (file, indent+XML_NESTED_INDENT, next);
    print_name_qualifier_end_tag (file, indent);
    }
  else
    {
    print_qualified_name_empty_tag (file, indent, DECL_NAME (decl));
    }
}


/* Given any _DECL, output the tags of the fully qualified form of its
   name.  */
void
xml_output_qualified_name (FILE* file, unsigned long indent, tree t)
{
  tree context;
  tree qualifiers = NULL_TREE;
  
  /* Build a list of all the qualifying contexts out to the global
     namespace.  */
  context = t;
  while (context && (context != global_namespace))
    {
    if (TYPE_P (context)) context = TYPE_NAME (context);

    qualifiers = tree_cons (NULL_TREE, context, qualifiers);

    context = CP_DECL_CONTEXT (context);
    }
  
  /* Output the qualifiers in the list.  */
  if (qualifiers)
    output_qualifier_list (file, indent, qualifiers);
}


/* Convert the identifier IN_ID to an XML encoded form by passing its string
   to xml_get_encoded_string_from_string().  */
const char*
xml_get_encoded_string (tree in_id)
{
  if(in_id)
    return xml_get_encoded_string_from_string (IDENTIFIER_POINTER (in_id));
  else
    return "";
}

#define XML_AMPERSAND     "&amp;"
#define XML_LESS_THAN     "&lt;"
#define XML_GREATER_THAN  "&gt;"
#define XML_SINGLE_QUOTE  "&apos;"
#define XML_DOUBLE_QUOTE  "&quot;"

/* Convert the name IN_STR to an XML encoded form.
   This replaces '&', '<', and '>'
   with their corresponding unicode character references.  */
const char*
xml_get_encoded_string_from_string (const char* in_str)
{
  int length_increase = 0;
  const char* inCh;
  char* outCh;
  char* newStr;
  tree id;

  /* Count special characters and calculate extra length needed for encoded
     form of string.  */
  for(inCh = in_str; *inCh != '\0' ; ++inCh)
    {
    switch (*inCh)
      {
      case '&': length_increase += strlen(XML_AMPERSAND)-1; break;
      case '<': length_increase += strlen(XML_LESS_THAN)-1; break;
      case '>': length_increase += strlen(XML_GREATER_THAN)-1; break;
      case '\'': length_increase += strlen(XML_SINGLE_QUOTE)-1; break;
      case '"': length_increase += strlen(XML_DOUBLE_QUOTE)-1; break;
      }
    }

  newStr = (char*) malloc (strlen (in_str) + length_increase + 1);
  outCh = newStr;

  /* Create encoded form of string.  */
  for(inCh = in_str ; *inCh != '\0' ; ++inCh)
    {
    switch (*inCh)
      {
      case '&': strcpy (outCh, XML_AMPERSAND); outCh += strlen(XML_AMPERSAND); break;
      case '<': strcpy (outCh, XML_LESS_THAN);  outCh += strlen(XML_LESS_THAN); break;
      case '>': strcpy (outCh, XML_GREATER_THAN);  outCh += strlen(XML_GREATER_THAN); break;
      case '\'': strcpy (outCh, XML_SINGLE_QUOTE);  outCh += strlen(XML_SINGLE_QUOTE); break;
      case '"': strcpy (outCh, XML_DOUBLE_QUOTE);  outCh += strlen(XML_DOUBLE_QUOTE); break;
      default: *outCh++ = *inCh;
      }
    }  

  *outCh = '\0';

  /* Ask for "identifier" of this string and free our own copy of the
     memory.  */
  id = get_identifier(newStr);
  free(newStr);
  return IDENTIFIER_POINTER (id);
}

#undef XML_AMPERSAND
#undef XML_LESS_THAN
#undef XML_GREATER_THAN
#undef XML_SINGLE_QUOTE
#undef XML_DOUBLE_QUOTE

/* Print N spaces to FILE.  Used for indentation of XML output.  */
void
print_indent (FILE* file, unsigned long n)
{
  unsigned long left = n;
  while(left >= 10)
    {
    fprintf(file, "          ");
    left -= 10;
    }
  while(left > 0)
    {
    fprintf(file, " ");
    left -= 1;
    }
}


/* Lookup the real name of an operator whose ansi_opname or ansi_assopname
   is NAME.  */
tree
reverse_opname_lookup (tree name)
{
  static const char* unknown_operator = "{unknown operator}";
  int i;

  /* Make sure we know about this internal name.  */
  if (!IDENTIFIER_OPNAME_P (name))
    return get_identifier (unknown_operator);

  /* Search the list of internal anmes */
  for (i=0; i < LAST_CPLUS_TREE_CODE ; ++i)
    {
    if (ansi_opname(i) == name)
      return get_identifier (operator_name_info[i].name);
    else if (ansi_assopname(i) == name)
      return get_identifier (assignment_operator_name_info[i].name);
    }
  
  return get_identifier (unknown_operator);
}

